/*
 * Copyright (c) 2007 Sun Microsystems, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/* STATUS
 * 
 * Coded sending but not tested
 * 
 * ToDo: 
 * - close connections on the stop
 * - add specific message code to the broadcasted packets
 * 
 */

package org.sunspotworld.demo;

import com.sun.spot.service.IService;
import com.sun.spot.util.Utils;

import java.io.IOException;
import javax.microedition.io.Connector;
import javax.microedition.io.Datagram;
import javax.microedition.io.DatagramConnection;

/**
 * Routines to control and read data from the SPOT's accelerometer and send them
 * via Radiogram packets to a display program running on a host computer.
 *<p>
 * The actual task scheduling is done by the PeriodicTask parent class.
 *<p>
 * Packets from the host application are received by the PacketReceiver class and
 * then dispatched to this class to respond to. The replies are sent back via the
 * PacketTransmitter class. The commands are defined in the PacketTypes class.
 * <p>
 * @author Ron Goldman<br>
 * Date: May 8, 2006, revised: August 1, 2007
 *
 * @see PeriodicTask
 * @see PacketHandler
 * @see PacketReceiver
 * @see PacketTransmitter
 * @see PacketTypes
 */
public class Broadcaster  
implements IService{
    
    protected int status = STOPPED;
    private int sampleInterval;     // in milliseconds
    
    protected Thread thread = null;
    
    // Location as int
    private static final int HOME = 1;  
    private static final int OFFICE = 2;  
    private static final int CORRIDOR = 3;  
    private static final int OUTDOOR = 4;  

    public static final byte LOCATE_INFORMATION_ID  = 1;
    
    public static final int MY_LOCATION = OFFICE;
    
    // datagram objects
    DatagramConnection dgConnection = null;
    Datagram dg = null;

    
    /**
     * Create a new accelerometer controller.
     *
     * @param m reference to the main program getting commands from the host display
     * @param sampleInterval how often to sample the accelerometer, in milliseconds
     */
    public Broadcaster(int sampleInterval) {
        
        this.sampleInterval = sampleInterval;
        System.out.println("Broadcaster Constructor Started");
        start(); // starting to send the packages
    }
    
    
    
    
    ///////////////////////////////////
    //
    // PeriodicTask overridden methods
    //
    ///////////////////////////////////
    
    /**
     * Routine called when task execution is about to start up.
     */
    public void starting() {
        System.out.println("Starting broadcaster sampler");
        
        // We create a DatagramConnection
        try {
            // The Connection is a broadcast so we specify it in the creation string
            dgConnection = (DatagramConnection) Connector.open("radiogram://broadcast:37");
            // Then, we ask for a datagram with the maximum size allowed
            dg = dgConnection.newDatagram(dgConnection.getMaximumLength());
        } catch (IOException ex) {
            System.out.println("Could not open radiogram broadcast connection");
            ex.printStackTrace();
            return;
        }


        
    }
    
    /**
     * Routine called when task execution is finished.
     */
    public void stopping() {
        System.out.println("Stopping broadcaster sampler");
    }


    /**
     * Called once per task 
     */
    public void doTask() {
        starting();
        
        status = RUNNING;
        while(status == RUNNING && thread == Thread.currentThread()){
            try {
                dg.reset();
                dg.writeByte(LOCATE_INFORMATION_ID);
                dg.writeInt(MY_LOCATION);
                dgConnection.send(dg);
                //System.out.println("Broadcast is going through");


            }catch (IOException ex) {
                   ex.printStackTrace(); 
            }
            Utils.sleep(this.sampleInterval);
        }
    }

     public boolean stop() {
        if (status != STOPPED) {
            status = STOPPING;
        }
        return true;
    }

    /**
     * Start periodic task execution.
     *
     * @return true if will start running
     */
    public boolean start() {
        if (status == STOPPED || status == STOPPING) {
            status = STARTING;
            thread = new Thread() {
                public void run() {
                    doTask();
                }
            };
            thread.start();
            Thread.yield();
        }
        return true;
    }

    /**
     * Pause the service, and return whether successful.
     *
     * Since there is no particular state associated with this service
     * then pause() can be implemented by calling stop().
     *
     * @return true if the service was successfully paused
     */
    public boolean pause() {
        return stop();
    }

    /**
     * Resume the service, and return whether successful.
     *
     * Since there was no particular state associated with this service
     * then resume() can be implemented by calling start().
     *
     * @return true if the service was successfully resumed
     */
    public boolean resume() {
        return start();
    }

    /**
     * Return service name
     *
     * @return the name of this service
     */
    public String getServiceName() {
        return "";
    }

    /**
     * Assign a name to this service.
     *
     * @param who the new name for this service
     */
    public void setServiceName(String who) {
        if (who != null) {
            //name = who;
        }
    }

    /**
     * Return if currently running.
     *
     * @return true if currently running
     */
    public boolean isRunning() {
        return status == RUNNING;
    }

    /**
     * Return current task execution status.
     *
     * @return current task execution status: STOPPED, STARTING, RUNNING, or STOPPING.
     */
    public int getStatus() {
        return status;
    }

    /**
     * Return whether service is started automatically on reboot.
     *
     * @return false as this service is never started automatically on reboot
     */
    public boolean getEnabled() {
        return false;
    }

    /**
     * Enable/disable whether service is started automatically. Noop for task execution.
     *
     * @param enable ignored
     */
    public void setEnabled(boolean enable) {
        // ignore
    }    

    
    
    
}


