/*
 * Copyright (c) 2007 Sun Microsystems, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/* STATUS
 * 
 * Aquisition and threading working
 * "movement" and "lack of movement" being detected (altough the algorithm is not good)
 * 
 * ToDo: 
 * - implement offsets
 * - passing this info as event
 * 
 */

/* the measure of movement or lack of movement is being based on the deviation or conformance
 * to the G value of the idle sunspot for a period of X samplings
 * where X can be adjusted and G (corresponding to the intensity of the acceleration vector) also
 * 
 * as expected this algorithm is not really much good once that if the spots are held differently
 * the pressing will be applied to more or less axes and change the value
 */  

package org.sunspotworld;

import org.sunspotworld.demo.util.*;


import com.sun.spot.sensorboard.EDemoBoard;
import com.sun.spot.sensorboard.peripheral.LIS3L02AQAccelerometer;

import java.io.IOException;

/**
 * Routines to control and read data from the SPOT's accelerometer and send them
 * via Radiogram packets to a display program running on a host computer.
 *<p>
 * The actual task scheduling is done by the PeriodicTask parent class.
 *<p>
 * Packets from the host application are received by the PacketReceiver class and
 * then dispatched to this class to respond to. The replies are sent back via the
 * PacketTransmitter class. The commands are defined in the PacketTypes class.
 * <p>
 * @author Ron Goldman<br>
 * Date: May 8, 2006, revised: August 1, 2007
 *
 * @see PeriodicTask
 * @see PacketHandler
 * @see PacketReceiver
 * @see PacketTransmitter
 * @see PacketTypes
 */
public class AccelMonitorTest extends PeriodicTask {
    
    // Definitions for raw accelerometer packets
    private static final double ACC_TRESHOLD_VARIANCE = 30;   // time (long) + num samples (byte)
    private static final int ACC_SAMPLING_TRESHOLD = 2;   // how many times I should detect that it is moving (or idle) before sayint that it is
    
    private LIS3L02AQAccelerometer acc;
    private int index = 0;

    private StartApplication main;
    private int sampleInterval;     // in milliseconds
    
    
    private double x;
    private double y;
    private double z;
    private double g; // intensity of the vector; sqrt of the sum of the modules

    private double gTreshold; // if is more or less then this plus the variance, it is considered moving
    private double variance = ACC_TRESHOLD_VARIANCE;
    private int nbSamples = 0; // number of samples that consist
    private boolean moving = false; // this variable keeps track of the state if
    // it is moving or stopped
    
    
    /**
     * Create a new accelerometer controller.
     *
     * @param m reference to the main program getting commands from the host display
     * @param sampleInterval how often to sample the accelerometer, in milliseconds
     */
    public AccelMonitorTest(StartApplication m, int sampleInterval) {
        super(3, sampleInterval, Thread.MAX_PRIORITY);
        this.sampleInterval = sampleInterval;
        main = m;
        acc = (LIS3L02AQAccelerometer)EDemoBoard.getInstance().getAccelerometer();
        acc.setScale(LIS3L02AQAccelerometer.SCALE_2G);        // start using 2G scale
        index = 0;
     
        start();
    }
    
    
    
    ///////////////////////////////////
    //
    // PeriodicTask overridden methods
    //
    ///////////////////////////////////
    
    /**
     * Routine called when task execution is about to start up.
     */
    public void starting() {
        System.out.println("Starting accelerometer sampler");
        // aquiring my trashold (ideally a proper calibration should be performed)
        try {
            x = acc.getAccelX();
            y = acc.getRawY();
            z = acc.getRawZ();
            g = Math.sqrt(x*x + y*y + z*z );
            gTreshold = g; // TRESHOLD SET!!!
        } catch (IOException ie) {
            System.out.println("IO exception: " + ie.toString());
        }

        
    }
    
    /**
     * Routine called when task execution is finished.
     */
    public void stopping() {
        System.out.println("Stopping accelerometer sampler");
    }


    /**
     * Called once per task 
     */
    public void doTask() {
        try {
            x = acc.getAccelX();
            y = acc.getRawY();
            z = acc.getRawZ();
            g = Math.sqrt(x*x + y*y + z*z );
            
            //System.out.println("Time = " + ((int) (System.currentTimeMillis() - startTime)) 
            //       + " and values of X, Y, Z, G are " + x + " "
            //       + y + " " + g + " ");
            
            // The logic is it check if there is change of state (movement over stopped or stopped over movement)
            // for ACC_SAMPLING_TRESHOLD consecutive times, if there is I change the state and send an event
            
            if(moving){
                if( (g<(gTreshold+variance)) &&
                    (g>(gTreshold-variance)) )
                    nbSamples++; // it means it stopeed                
                else
                    nbSamples = 0; // if it is not consecutive I restart the counter
            }
            else{ // stoped
                if( (g>(gTreshold+variance)) ||
                    (g<(gTreshold-variance)) )
                    nbSamples++; // it means it moved
                else
                    nbSamples = 0; // if it is not consecutive I restart the counter
            }
            
            if (nbSamples > ACC_SAMPLING_TRESHOLD){ // I assume that it really changed the state
                moving = !moving;
                nbSamples = 0;
                main.receiveAccelInternalSignal(moving);
                //System.out.println("E diga silvio, ele se move? " + moving);
            }
            

        } catch (IOException ie) {
            System.out.println("IO exception: " + ie.toString());
        }
    }

    /** temporary fix until IService interface fixed */
    public void setName(String who){};
}
